"""Custom implementation of Rule LT02 to allow indents and new lines after
create statements.
"""

from typing import List

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.utils.reflow.sequence import ReflowSequence


class Rule_EasyQL_LT02(BaseRule):
    """Incorrect Indentation.

    This is a modified version of the core sqlfluff rule LT02.

    **Anti-pattern**

    The ``•`` character represents a space and the ``→`` character represents
    a tab. In this example, the third line contains five spaces instead of four
    and the second line contains two spaces and one tab.

    .. code-block:: sql
       :force:

        SELECT
        ••→a,
        •••••b
        FROM foo


    **Best practice**

    Change the indentation to use a multiple of four spaces. This example also
    assumes that the ``indent_unit`` config value is set to ``space``. If it
    had instead been set to ``tab``, then the indents would be tabs instead.

    .. code-block:: sql
       :force:

        SELECT
        ••••a,
        ••••b
        FROM foo

    """

    name = "layout.indent"
    # NOTE: We're combining three legacy rules here into one.
    aliases = ("L002", "L003", "L004")
    groups = ("all", "core", "layout")
    crawl_behaviour = RootOnlyCrawler()
    is_fix_compatible = True
    targets_templated = True
    template_safe_fixes = True
    _adjust_anchors = True

    def handle_create_layout(self) -> List[LintResult]:
        # return an empty list because the indentation in create statements
        # works differently
        return list()

    def _eval(self, context: RuleContext) -> List[LintResult]:
        """Indentation not consistent with previous lines.

        To set the default tab size, set the `tab_space_size` value
        in the appropriate configuration. To correct indents to tabs
        use the `indent_unit` value set to `tab`.

        """
        reflows: List[LintResult] = list()

        for sub_segment in context.segment.segments:
            if len(sub_segment.segments) > 0 and \
                    "create" in sub_segment.segments[0].type:
                reflows.extend(self.handle_create_layout())
            else:
                reflows.extend(
                    ReflowSequence.from_around_target(
                        sub_segment,
                        context.segment,
                        context.config)
                    .reindent()
                    .get_results()
                )

        return reflows



"""An example of a custom rule implemented through the plugin system.

This uses the rules API supported from 0.4.0 onwards.
"""

from sqlfluff.core.rules import (
    BaseRule,
    LintResult,
    RuleContext
)
from sqlfluff.utils.functional import FunctionalContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_EasyQL_L002(BaseRule):
    """JOIN and ON keyword should be in the same line.
    """
    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler({"join_clause"})

    def __init__(self, *args, **kwargs):
        """Overwrite __init__ to set config."""
        super().__init__(*args, **kwargs)

    def _eval(self, context: RuleContext):
        """We should not JOIN .. ON differnet lines."""
        on_appeared = False
        for seg in context.segment.segments:
            if seg.raw.lower() == "on":
                on_appeared = True
            if seg.raw == "\n" and not on_appeared:
                desc = "Join and ON keyword should be at the same line."
                return LintResult(
                    anchor=seg,
                    description=desc
                )


class Rule_EasyQL_L003(BaseRule):
    """Prohibit the use of 'WHERE 1=1' in queries."""
    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler({"where_clause"})
    is_fix_compatible = False

    def _eval(self, context: RuleContext):
        text_segments = (
            FunctionalContext(context)
            .segment.children()
        )

        for idx, seg in enumerate(text_segments):
            # Look for the pattern '1=1'
            if "1=1" in seg.raw_upper.replace(" ", ""):
                return LintResult(anchor=seg)
        return None



from typing import List, Type

from sqlfluff.core.config import ConfigLoader
from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule


@hookimpl
def get_rules() -> List[Type[BaseRule]]:
    """Get plugin rules.

    NOTE: It is much better that we only import the rule on demand.
    The root module of the plugin (i.e. this file which contains
    all of the hook implementations) should have fully loaded before
    we try and import the rules. This is partly for performance
    reasons - but more because the definition of a BaseRule requires
    that all of the get_configs_info() methods have both been
    defined _and have run_ before so all the validation information
    is available for the validation steps in the meta class.
    """
    # i.e. we DO recommend importing here:
    from sqlfluff_easy_ql.rules import (
        Rule_EasyQL_L002,
        Rule_EasyQL_L003
    )  # noqa: F811
    from sqlfluff_easy_ql.LT02 import Rule_EasyQL_LT02
    from sqlfluff_easy_ql.LT01 import Rule_EasyQL_LT01

    return [Rule_EasyQL_LT01, Rule_EasyQL_LT02,
            Rule_EasyQL_L002, Rule_EasyQL_L003]


@hookimpl
def load_default_config() -> dict:
    """Loads the default configuration for the plugin."""
    return ConfigLoader.get_global().load_config_resource(
        package="sqlfluff_easy_ql",
        file_name="easy_ql_default_config.cfg",
    )


@hookimpl
def get_configs_info() -> dict:
    """Get rule config validations and descriptions."""
    return {
        "forbidden_columns": {"definition": "A list of column to forbid"},
    }



"""Implementation of custom rule for indentation in CREATE statements."""

from typing import List

from sqlfluff.core.parser import NewlineSegment, WhitespaceSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.parser.segments.base import BaseSegment
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_EasyQL_LT01(BaseRule):
    """The name of the objects created in CREATE statements should be indented.
    """
    groups = ("all",)
    crawl_behaviour = SegmentSeekerCrawler({
        # more or less all matches to create_(.*)_statement
        "create_table_statement",
        "create_index_statement",
        "create_view_statement",
        "create_function_statement",
        "create_procedure_statement"
    })
    is_fix_compatible = True

    def _aux_fix(
            self,
            segments,
            object_name_segment,
            object_name_idx
    ) -> List[LintFix]:
        """Applies the fix logic for the different possible cases."""
        fix_segments: List[BaseSegment] = list()

        # check whether there is a newline
        if segments[object_name_idx-2].type != "newline" and \
                segments[object_name_idx-1].type != "newline":
            fix_segments.append(NewlineSegment())
        # check the indentation
        if segments[object_name_idx-1].type != "whitespace" or \
                len(segments[object_name_idx-1].raw) != 4:
            fix_segments.append(WhitespaceSegment("    "))

        fix = LintFix.create_before(
            object_name_segment,
            fix_segments
        )

        if segments[object_name_idx-1].type == "whitespace" and \
                len(segments[object_name_idx-1].raw) == 1:
            fix2 = LintFix.delete(segments[object_name_idx-1])
            return [fix, fix2]

        return [fix]

    def _eval(self, context: RuleContext) -> List[LintResult]:
        # skip entire part of CREATE [OR REPLACE]
        segments = context.segment.segments
        obj_reference = context.segment.type.split("_")[1] + "_reference"
        table_name_idx, table_name_segment = next(
            ((idx, s) for idx, s in enumerate(segments)
                if s.type == obj_reference or s.type == "function_name")
        )

        # assert that there is a newline and 4 spaces before the name
        if segments[table_name_idx-2].type == "newline" and \
                segments[table_name_idx-1].type == "whitespace" and \
                len(segments[table_name_idx-1].raw) == 4:
            return list()
        else:  # apply fixes and return lint result
            fixes_to_apply = self._aux_fix(
                segments, table_name_segment, table_name_idx
            )

            return [
                LintResult(
                    anchor=table_name_segment,
                    fixes=fixes_to_apply,
                    description="The name of the created object must be in a new line and indented."  # noqa: E501
                )
            ]
